import numpy

from dartqc.DartReader import DartReader
from dartqc.DartUtils import stamp


class Preprocessor(DartReader):

    def __init__(self, call_data, call_attributes):

        """
        Give the Preprocessor the call data and inherit from DartReader. Use the attributes of DartReader
        (self.data, self.attributes) to process the read count matrix and collapse replicate read counts for samples
        present in the call data.
        """

        DartReader.__init__(self)

        self.name = "preprocessor"

        self.call_data = call_data
        self.call_attributes = call_attributes

        self.call_names = call_attributes["sample_names"]

        self.replicates = {}

        self._set_log()

    def _set_log(self):

        self.call_attributes["modules"][self.name] = {
            "results": {},
            "settings": {},
            "states": {}  # States are other parameters of interest not necessary results or settings.
        }

    def get_replicates(self):

        """
        Get replicate indices across individuals for summation of read calls.

        """

        for i, sample in enumerate(self.sample_names):
            if sample not in self.replicates.keys():
                self.replicates[sample] = [i]
            else:
                self.replicates[sample] += [i]

    def read_count_data(self, file):

        """
        Alternative call to reading double row format for inputting and transforming the read count matrix,
        read the data without encoding, i.e. in list of tuple format for read counts per call.
        """

        self.read_double_row(file=file, encode=False, numeric=True)

    def check_concordance(self):

        if set(self.sample_names) != set(self.call_names):
            stamp("Sample names from the read count file are not the same as sample names from the data file.")
            stamp("Sample difference, present in one but not the other data:")
            for sample in set(self.sample_names).difference(set(self.call_names)):
                stamp(sample)
            exit(0)
        else:
            stamp("Concordance between sample names in call and count data, all is good.")

        if len(self.call_data) != len(self.data):

            diff = set(self.call_data.keys()).difference(set(self.data.keys()))
            inter = set(self.call_data.keys()).intersection(set(self.data.keys()))

            stamp("Number of SNPs are different, there are:", len(self.call_data), "SNPs in the called set and",
                  len(self.data), "SNPs in the raw set.")

            stamp(len(diff), "SNPs have a different ID. Keeping the intersection of", len(inter), "SNPs...")

            self.data = {k: v for (k, v) in self.data.items() if k in inter}
            self.call_data = {k: v for (k, v) in self.call_data.items() if k in inter}

        if set(self.call_data.keys()) != set(self.data.keys()):
            stamp("SNP IDs are not the same, removal not effective, please re-format your data.")

    def get_missing(self):

        """

        Get count of missing alleles in call data, to subtract from total / replaced in filtering read counts

        """

        missing = 0

        for k, v in self.call_data.items():
            missing += v["calls"].count("-")

        return missing

    def filter_read_counts(self, threshold=7):

        """
        1. Transform read count matrix to numpy array, ordered by allele IDs.
        2. Sum-collapse replicate columns in the order of sample names from the original data (sample_names)
        3. Construct the reduced array and assign each call in a dictionary the allele ID
        4. For each allele in the dictionary, construct a boolean vector with True if the sum of the two allele counts
           is smaller than the threshold value, otherwise False
        5. Use this vector in the same iteration to assign missing to all calls in the original data

        """

        self.check_concordance()
        call_missing = self.get_missing()
        stamp("Number of missing in call data:", call_missing)

        snp_order = sorted(self.data.keys())
        reduced_counts = {}

        stamp("Finding replicate columns...")

        self.get_replicates()

        stamp("Ordering count data by SNPs...")

        counts = [self.data[snp]["calls"] for snp in snp_order]

        count_array = numpy.asarray(counts)

        stamp("Sum-collapsing replicates...")

        columns = [numpy.sum(count_array[:, self.replicates[sample]], axis=1).tolist() for sample in self.call_names]

        reduced_array = list(zip(*columns))

        for i, snp in enumerate(snp_order):
            reduced_counts[snp] = reduced_array[i]

        replaced = 0
        total = 0

        stamp("Replacing low counts with missing...")

        for snp, counts in reduced_counts.items():
            filter_vector = [False if sum(allele_counts) <= threshold else True for allele_counts in counts]
            self.call_data[snp]["calls"] = [call if filter_vector[i] else "-" for i, call in
                                            enumerate(self.call_data[snp]["calls"])]

            replaced += filter_vector.count(False)
            total += len(filter_vector)

        replaced -= call_missing

        stamp("Pre-processing silenced {r}/{t} calls {p}".format(r=replaced, t=total,
                                                                 p=format((replaced/total)*100, ".2f")))

        self.call_attributes["modules"][self.name]["results"] = {
            "total_calls": total,
            "replaced_calls": replaced,
            "before_missing": call_missing,
            "after_missing": call_missing + replaced
        }

        self.call_attributes["modules"][self.name]["settings"] = {
            "read_count_sum_threshold": threshold
        }

    def get_data(self):

        return self.call_data, self.call_attributes



import os
import time
import argparse
import textwrap

from subprocess import call, check_output, CalledProcessError


class Installer:

    def __init__(self, miniconda=True):

        self.miniconda = miniconda
        self.miniconda_url = "https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86_64.sh"
        self.miniconda_install = "Miniconda3-latest-Linux-x86_64.sh"

        self.base_path = os.path.dirname(os.path.realpath(__file__))

        self.env = os.path.join(self.base_path, "env", "dartqc.yaml")

        if not self._check() and self.miniconda:
            self._install_miniconda()

        self._install_env()

    @staticmethod
    def _check():

        try:
            check_output(["conda", "--version"])
            stamp("Conda manager detected, use: conda --version")
            return True
        except CalledProcessError:
            stamp("Could not detect conda manager, please see README.")
            return False

    def _install_miniconda(self):

        stamp("Could not detect Conda. Installing Miniconda for Python 3.")

        try:
            stamp("Downloading and installing...")
            with open("install.log", "w") as err_file:
                check_output(["wget", self.miniconda_url, "&&", "bash", self.miniconda_install,
                              "-b", "-p", "$HOME/miniconda"], stderr=err_file)
                stamp("Success. Removing installer for Miniconda...")
                os.remove(os.path.join(os.getcwd(), self.miniconda_install))
                if os.path.exists("install.log"):
                    os.remove("install.log")
        except CalledProcessError:
            stamp("Could not install Miniconda, please see install.log and README.")

        try:
            stamp("Adding $HOME/miniconda to PATH.")
            with open("install.log", "w") as err_file:
                call("""echo 'export PATH="$HOME/miniconda/bin:$PATH"' >> $HOME/.bashrc && source .bashrc""",
                     stderr=err_file)
                stamp("Done. Testing...")
                self._check()
                if os.path.exists("install.log"):
                    os.remove("install.log")
        except CalledProcessError:
            stamp("Could not install Miniconda, please see install.log and README.")

    def _install_env(self):

        try:
            with open("install.log", "w") as err_file:
                stamp("Installing environment for DartQC...")
                check_output(["conda", "env", "create", "--name", "dartqc", "--file", self.env], stderr=err_file)
            stamp("Installed environment, activate with: source activate dartqc")
            if os.path.exists("install.log"):
                os.remove("install.log")
        except CalledProcessError:
            stamp("Could not install environment, please see install.log and README.")


class PBS:

    def __init__(self, walltime="02:00:00", email="user@hpc.jcu.edu.au"):

        self.walltime = walltime
        self.email = email

        self._write_pbs()

    def _write_pbs(self):

        pbs = textwrap.dedent("""
        #PBS -S /bin/sh
        #PBS -M {email}
        #PBS -m n
        #PBS -j oe
        #PBS -l ncpus=1
        #PBS -l walltime={walltime}

        source ~/.bashrc
        source activate dartqc
        cd $PBS_O_WORKDIR

        echo $PATH

        if [ ! "$command" ]
            then
                echo -e "Error: script must be supplied with a command"
                exit 1
            else
                echo "$command"
                eval "$command"
        fi

        exit 0

        """.format(walltime=self.walltime, email=self.email))

        with open("dartqc.pbs", "w") as outfile:
            outfile.write(pbs)


class CommandLine:

    def __init__(self):

        parser = argparse.ArgumentParser()

        parser.add_argument("--version", "-v", dest="version", action="store_true", help="print version and exit")

        parser.add_argument("--output_path", "-o", type=lambda p: os.path.abspath(p),
                            default=os.getcwd(), required=False, dest="out_path", help="output path")

        parser.add_argument("--project", "-p", type=str, default="dartqc", required=False,
                            dest="project", help="project name")

        parser.add_argument("--populations", "--pop", type=lambda p: os.path.abspath(p), default=None, required=False,
                            dest="pop_file", help="CSV file with header columns ID and Population")

        subparsers = parser.add_subparsers(help='Command-line interface for DartQC')

        install_parser = subparsers.add_parser("install")

        install_parser.set_defaults(subparser='install')

        pbs_parser = subparsers.add_parser("pbs")

        pbs_parser.add_argument("--email", "-e", type=str, default="user@hpc.jcu.edu.au", required=False,
                                dest="email", help="user email for pbs")

        pbs_parser.add_argument("--walltime", "-w", type=str, default="02:00:00", required=False,
                                dest="walltime", help="expected walltime for pbs")

        pbs_parser.set_defaults(subparser='pbs')

        prepare_parser = subparsers.add_parser("prepare")

        prepare_parser.add_argument("--file", "-f", type=lambda p: os.path.abspath(p), required=True,
                                    dest="file", help="path to input file")
        prepare_parser.add_argument("--name", "-n", type=str, default=None, required=False,
                                    dest="output_name", help="name of output scheme file")
        prepare_parser.add_argument("--sheet", "-s", type=str, default=None, required=False,
                                    dest="sheet", help="if file is excel: name of sheet that contains double row data")

        prepare_parser.set_defaults(subparser='prepare')

        process_parser = subparsers.add_parser("process")

        process_parser.add_argument("--raw", "-r", type=lambda p: os.path.abspath(p), required=True,
                                    dest="raw_file", help="path to raw read file")

        process_parser.add_argument("--raw_scheme", default="raw_scheme.json",
                                    type=lambda p: os.path.abspath(p), required=False,
                                    dest="raw_scheme", help="path to raw scheme json file")

        process_parser.add_argument("--read_sum", "--reads", default=10, type=int, required=False,
                                    dest="raw_read_threshold",
                                    help="silence call if ref and snp allele raw read sum < threshold")

        process_parser.add_argument("--calls", "-c", default="calls.csv", type=lambda p: os.path.abspath(p),
                                    required=True, dest="call_file", help="path to called read file")

        process_parser.add_argument("--call_scheme", default="call_scheme.json",
                                    type=lambda p: os.path.abspath(p), required=False,
                                    dest="call_scheme", help="path to call scheme json file")

        process_parser.set_defaults(subparser='process')

        filter_parser = subparsers.add_parser("filter")

        filter_parser.add_argument("--processed", "--pp", type=lambda p: os.path.abspath(p), required=False,
                                   dest="processed_path", default=None,
                                   help="input path to processed data files (project_data.json, project_attr.json)")

        filter_parser.add_argument("--calls", "-c", default="calls.csv", type=lambda p: os.path.abspath(p),
                                   required=False, dest="call_file", help="path to called read file")
        filter_parser.add_argument("--call_scheme", default="call_scheme.json",
                                   type=lambda p: os.path.abspath(p), required=False,
                                   dest="call_scheme", help="path to call scheme json file")

        filter_parser.add_argument("--maf", default=[None], type=lambda s: [float(item) for item in s.split(',')],
                                   dest="maf", help="filter snps <= minor allele frequency")
        filter_parser.add_argument("--hwe", default=[None], type=lambda s: [float(item) for item in s.split(',')],
                                   dest="hwe", help="filter snps <= p-value of hardy-weinberg test")
        filter_parser.add_argument("--call_rate", default=[None], type=lambda s: [float(item) for item in s.split(',')],
                                   dest="call_rate", help="filter snps <= call rate of snp")
        filter_parser.add_argument("--rep", default=[None], type=lambda s: [float(item) for item in s.split(',')],
                                   dest="rep", help="filter snps <= replication average of snp")

        filter_parser.add_argument("--mind", default=[None], type=lambda s: [float(item) for item in s.split(',')],
                                   dest="mind", help="filter samples > missingness per sample")
        filter_parser.add_argument("--mono", default=None,
                                   dest="mono", help="filter samples monomorphic in <mono> populations ('all', int)")
        filter_parser.add_argument("--mono_comparison", default="==",
                                   dest="mono_comp", help="filter samples monomorphic in >=, <=, == populations ('==')")

        filter_parser.add_argument("--split_clones", default="", type=str, dest="split_clones",
                                   help="split clone ids on this character")

        filter_parser.add_argument("--duplicates", default=False, action="store_true",
                                   dest="remove_duplicates", help="remove snps with duplicate clone IDs")
        filter_parser.add_argument("--clusters", default=False, action="store_true",
                                   dest="remove_clusters", help="remove snps in identical sequence clusters")
        filter_parser.add_argument("--identity", default=0.95, type=float,
                                   dest="identity", help="remove snps in identical sequence clusters")

        filter_parser.set_defaults(subparser='filter')

        self.args = parser.parse_args()


def stamp(*args):

    print(str(time.strftime("[%H:%M:%S]")) + " " + " ".join([str(arg) for arg in args]))


import os
import csv
import numpy
import json

from dartqc.DartUtils import stamp


class DartWriter:

    def __init__(self, data, attributes):

        self.data = data
        self.attributes = attributes

        self.decoding_scheme = dict()

        self.set_encoding()

    def set_encoding(self, homozygous_major=("A", "A"), homozygous_minor=("B", "B"), heterozygous=("A", "B"),
                     missing=("0", "0")):

        self.decoding_scheme = {"-": missing, "0": heterozygous, "1": homozygous_minor, "2": homozygous_major}

    def write_plink(self, file_name, sep="\t", remove_space=False):

        snp_order = sorted(self.data.keys())

        stamp("Decoding calls...")
        snp_rows = [[self.decoding_scheme[snp] for snp in self.data[snp_id]["calls"]] for snp_id in snp_order]

        stamp("Transposing calls...")
        snps_by_sample = numpy.asarray(snp_rows).transpose(1, 0, 2)

        genotypes = [sample.flatten().tolist() for sample in snps_by_sample]

        names = self.attributes["sample_names"]
        pops = [self.attributes["pops"][sample] for sample in names]

        if remove_space:
            names = ["_".join(name.split()) for name in names]
            pops = ["_".join(pop.split()) for pop in pops]

        ped_file = os.path.join(self.attributes["out_path"], file_name + '.ped')
        map_file = os.path.join(self.attributes["out_path"], file_name + '.map')

        paternal = ["0"] * len(names)
        maternal = ["0"] * len(names)
        sex = ["0"] * len(names)
        phenotype = ["-9"] * len(names)

        plink = zip(pops, names, paternal, maternal, sex, phenotype, genotypes)

        stamp("Formatting calls...")

        ped_data = []
        for row in plink:
            new_row = list(row[:6])
            for geno in row[6]:
                new_row.append(geno)
            ped_data.append(new_row)

        stamp("Writing PLINK")
        stamp("PED file:", ped_file)
        stamp("MAP file:", map_file)

        with open(ped_file, 'w') as ped_out:
            ped_writer = csv.writer(ped_out, delimiter=sep)
            ped_writer.writerows(ped_data)

        # MAP Formatting

        map_data = [["0", snp_id, "0", "0"] for snp_id in snp_order]

        with open(map_file, 'w') as map_out:
            ped_writer = csv.writer(map_out, delimiter=sep)
            ped_writer.writerows(map_data)

    def write_json(self, file_name, data_indent=0, attribute_indent=4):

        data_file = os.path.abspath(os.path.join(self.attributes["out_path"], file_name + "_data.json"))
        attribute_file = os.path.abspath(os.path.join(self.attributes["out_path"], file_name + "_attr.json"))

        stamp("Writing data to JSON")
        stamp("Data file:", data_file)
        stamp("Attribute file:", attribute_file)

        with open(data_file, "w") as data_out:
            json.dump(self.data, data_out, indent=data_indent)

        with open(attribute_file, "w") as attr_out:
            json.dump(self.attributes, attr_out, indent=attribute_indent)



import textwrap

from dartqc.DartUtils import stamp

class DartMessages:

    def __init__(self):

        pass

    def get_filter_message(self, filter, threshold, initial, removed, retained, time=True):

        filter_msg = textwrap.dedent("""
                SNP Filter
        -------------------------------

        {0} at {1}

        Initial:    {2}
        Removed:    {3}
        Retained:   {4}

        -------------------------------
        """ .format(filter.upper(), threshold, initial, removed, retained))

        if time:
            stamp("Filtered {0} at {1}".format(filter.upper(), threshold))
            stamp("Removed {0} SNPs".format(removed))
            stamp("Retained {0} SNPs".format(retained))
        else:
            print(filter_msg)

    def get_redundancy_message(self, type, initial, removed, retained, time=True):

        redundancy_msg = textwrap.dedent("""
                  REDUNDANCY
        -------------------------------

        {0}

        Initial:    {1}
        Removed:    {2}
        Retained:   {3}

        -------------------------------
        """ .format(type.upper(), initial, removed, retained))

        if time:
            stamp("Redundancy module {0}".format(type.upper()))
            stamp("Removed {0} SNPs".format(removed))
            stamp("Retained {0} SNPs".format(retained))
        else:
            print(redundancy_msg)

    def get_cdhit_message(self, identity, time=True):

        cluster_msg = textwrap.dedent("""
                  CLUSTERING
        -------------------------------

        Running CDHIT-EST...

        Threshold: {0}%

        -------------------------------
        """ .format(identity*100))

        if time:
            stamp("Running CD-HIT at nucleotide identity {0}%".format(identity*100))
        else:
            print(cluster_msg)


import csv
import operator
import os
import shutil
from subprocess import call

import pandas
import numpy
from Bio import SeqIO
from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from scipy import stats

from dartqc.DartUtils import stamp
from dartqc.DartMessages import DartMessages


class SummaryModule:

    def __init__(self, data=None, attributes=None, out_path=None):

        self.data = data
        self.attributes = attributes

        if out_path is None:
            self.out_path = attributes["out_path"]
        else:
            self.out_path = out_path

        os.makedirs(self.out_path, exist_ok=True)

    def write_snp_summary(self, file="snp_summary.csv", summary_parameters=None, sort=False):

        if summary_parameters is None:
            summary_parameters = ["maf", "hwe", "rep", "call_rate"]

        out_file = os.path.join(self.out_path, self.attributes["project"] + "_" + file)

        out_data = [["id"] + summary_parameters]

        snps = [[snp] + [data[parameter] for parameter in summary_parameters] for snp, data in self.data.items()]

        if sort:
            snps = sorted(snps, key=operator.itemgetter(*[i for i in range(1, len(summary_parameters)+1)]),
                          reverse=True)

        out_data += snps

        with open(out_file, "w") as snp_summary:
            writer = csv.writer(snp_summary)
            writer.writerows(out_data)

    def write_module_summary(self, file="module_summary.csv"):

        # Function at the moment for command line, hard-coded, need to fix.

        params_snp, removed_snp = self._get_snp_results()
        params_red, removed_red = self._get_redundancy_results()
        params_pop, removed_pop = self._get_pop_results()
        params_sam, removed_sam = self._get_sample_results()
        params_ppr, removed_ppr = self._get_preprocessing_results()

        project_param = {"project": self.attributes["project"], "snps": self.attributes["snps"]}

        snp_removed = self._get_snp_sum([removed_red, removed_pop, removed_snp])
        project_removed = {"project": self.attributes["project"], "snps": snp_removed}

        row_param = {k: v for d in [project_param, params_ppr, params_sam, params_pop, params_snp, params_red]
                     for k, v in d.items()}
        row_removed = {k: v for d in [project_removed, removed_ppr, removed_sam, removed_pop, removed_snp, removed_red]
                       for k, v in d.items()}

        df = pandas.DataFrame([row_param, row_removed], index=["parameters", "removed"])

        out_file = os.path.join(self.out_path, self.attributes["project"] + "_" + file)

        df.to_csv(out_file)

    @staticmethod
    def _get_snp_sum(dicts):

        s = 0
        for d in dicts:
            for k, v in d.items():
                if k in ("maf", "hwe", "rep_average", "monomorphic", "call_rate", "clusters", "duplicates"):
                    if v is not None:
                        s += int(v)

        return s

    def _get_snp_results(self):

        """Extract entry from Attributes"""

        try:
            results = self.attributes["modules"]["snp"]["results"]
            parameters = self.attributes["modules"]["snp"]["settings"]["parameters"]

            params = {entry[0]: entry[1] for entry in parameters}
            removed = {param: result["removed"] for param, result in results.items()}

        except KeyError:
            stamp("Could not detect results for SNP Module, skipping...")

            params = {"maf": None, "hwe": None, "call_rate": None, "rep_average": None}
            removed = {"maf": None, "hwe": None, "call_rate": None, "rep_average": None}

        return params, removed

    def _get_sample_results(self):

        try:
            removed = {"mind": self.attributes["individual"]["results"]["mind"]["removed_samples"],
                       "samples": self.attributes["individual"]["results"]["mind"]["removed_samples"]}

            params = {"mind": self.attributes["individual"]["results"]["mind"]["value"],
                      "samples": len(self.attributes["individual"]["states"]["mind"]["sample_names_original"])}

        except KeyError:
            stamp("Could not detect results for Sample Module, skipping...")

            params = {"mind": None, "samples": None}
            removed = {"mind": None, "samples": None}

        return params, removed

    def _get_pop_results(self):

        try:
            removed = {"monomorphic": self.attributes["modules"]["population"]["results"]["removed"]}
            params = {"monomorphic": self.attributes["modules"]["population"]["settings"]["value"]}
        except KeyError:
            stamp("Could not detect results for Population Module, skipping...")
            params = {"monomorphic": None}
            removed = {"monomorphic": None}

        return params, removed

    def _get_redundancy_results(self):

        try:
            parameters = self.attributes["modules"]["redundancy"]["settings"]
            params = {"clusters": parameters["clusters"], "duplicates": parameters["duplicates"],
                      "identity:": parameters["identity"]}
            results = self.attributes["modules"]["redundancy"]["results"]

            removed = {"clusters": results["clusters"]["removed"], "duplicates": results["duplicates"]["removed"],
                       "identity": None}
        except KeyError:
            stamp("Could not detect results for Redundancy Module, skipping...")
            params = {"clusters": None, "duplicates": None, "identity": None}
            removed = {"clusters": None, "duplicates": None, "identity": None}

        return params, removed

    def _get_preprocessing_results(self):

        try:
            params = {"preprocess": self.attributes["modules"]["preprocessor"]["settings"]["read_count_sum_threshold"],
                      "calls": self.attributes["modules"]["preprocessor"]["settings"]["results"]["total_calls"],
                      "missing": self.attributes["modules"]["preprocessor"]["settings"]["results"]["before_missing"]}
            removed = {"preprocess": self.attributes["modules"]["preprocessor"]["settings"]["results"]["replaced_calls"],
                       "calls": self.attributes["modules"]["preprocessor"]["settings"]["results"]["replaced_calls"],
                       "missing": self.attributes["modules"]["preprocessor"]["settings"]["results"]["replaced_calls"]}
        except KeyError:
            stamp("Could not detect results for Preprocessing Module, skipping...")
            params = {"preprocess": None, "calls": None, "missing": None}
            removed = {"preprocess": None, "calls": None, "missing": None}

        return params, removed

    def write_matrix(self, combination_matrix, r_matrix=None, file="combination_table.csv", r_file="r_matrix.csv"):

        out_file = os.path.join(self.out_path, file)

        with open(out_file, "w") as table_file:
            writer = csv.writer(table_file)
            writer.writerows(combination_matrix)

        if r_matrix is not None:
            out_r = os.path.join(self.out_path, r_file)
            with open(out_r, "w") as out_r_file:
                writer = csv.writer(out_r_file)
                writer.writerows(r_matrix)

########################################################################################################################


class QualityControl:

    def __init__(self, data, attributes):

        self.data = data                                # Dictionary holds data from DartReader
        self.attributes = attributes

        self.verbose = True

        self.messages = DartMessages()

        self.sample_size = attributes["sample_size"]
        self.sample_names = attributes["sample_names"]

        self.missing = attributes["missing"]
        self.homozygous_major = attributes["homozygous_major"]
        self.homozygous_minor = attributes["homozygous_minor"]
        self.heterozygous = attributes["heterozygous"]

        self.project = attributes["project"]
        self.out_path = attributes["out_path"]


########################################################################################################################


class PopulationModule(QualityControl):

    def __init__(self, data, attributes):

        QualityControl.__init__(self, data, attributes)

        self.name = "population"

        self.pops = attributes["pops"]
        self.sample_names = attributes["sample_names"]  # List of ordered unique names, same order as calls for SNPs

        self.populations = {}                           # Dictionary of populations and list of member indices
        self.monomorphics = {}                          # Dictionary of populations and list of mono SNPs for key pop

        self.attributes["modules"][self.name] = {}

        self._set_log()

        self._get_sample_indices()

    def _set_log(self):

        self.attributes["modules"][self.name] = {
            "results": {},
            "settings": {},
            "states": {}  # States are other parameters of interest not necessary results or settings.
        }

    def get_data(self, mono="all", comparison="=="):

        stamp("Initialised Population Module")

        if mono is None:
            stamp("No filter specified, returning data.")
            return self.data, self.attributes

        stamp("Indexing monomorphic SNPs in each population")
        self._calculate_monomorphics()

        for pop, indices in self.populations.items():
            stamp("There are", len(indices), "samples in population", pop)

        for pop, monomorphs in self.monomorphics.items():
            stamp("There are", len(monomorphs), "monomorphic SNPs in population", pop)

        # If threshold is string 'all', set to all populations.

        stamp("Filtering SNPs that are monomorphic in", mono, "populations.")

        if mono == "all":
            mono = len(self.populations)

        if comparison == "==":
            filtered = {snp: data for snp, data in self.data.items() if data["mono"] == mono}
        elif comparison == ">=":
            filtered = {snp: data for snp, data in self.data.items() if data["mono"] >= mono}
        elif comparison == "<=":
            filtered = {snp: data for snp, data in self.data.items() if data["mono"] <= mono}
        else:
            raise ValueError("Comparison must be one of: <=, >=, ==")

        filtered_data = {snp: data for snp, data in self.data.items() if snp not in filtered}

        stamp("Filtered", len(filtered), "SNPs.")

        attributes = self._log_monomorphic(self.attributes, filtered_data, mono)

        return filtered_data, attributes

    def _log_monomorphic(self, attributes, filtered_data, mono):

        attributes["modules"][self.name]["settings"] = {
            "parameter": "mono",
            "value": mono,
        }

        attributes["modules"][self.name]["results"] = {
            "before": len(self.data),
            "after": len(filtered_data),
            "removed": len(self.data) - len(filtered_data)
        }

        attributes["modules"][self.name]["states"] = {
            "monomorphic": self.monomorphics
        }

        return attributes

    def _get_sample_indices(self):

        """
        For each population get indices of sample names.

        """

        for name, pop in self.pops.items():
            name_index = self.sample_names.index(name)
            if pop not in self.populations:
                self.populations[pop] = [name_index]
            else:
                self.populations[pop].append(name_index)

    def _calculate_monomorphics(self):

        for pop in self.populations.keys():
            self.monomorphics[pop] = []

        for snp, data in self.data.items():
            number = 0
            for pop, indices in self.populations.items():

                # For each SNP get the unique set of calls that is not missing and belongs to the population (pop)
                calls = set([snp_call for i, snp_call in self._iterate_call_indices(data["calls"])
                             if i in indices and snp_call != self.missing])

                # If all calls in this set are the same (i.e. 1, 2 or 0) then
                # declare SNP monomorphic for this population

                if len(calls) == 1:
                    self.monomorphics[pop].append(snp)
                    number += 1

                # Repeat for each population

            # When done, add the total number of population the SNP
            # is monomorphic to the data for this SNP

            self.data[snp]["mono"] = number

    @staticmethod
    def _iterate_call_indices(calls):

        for i, snp_call in enumerate(calls):
            yield i, snp_call


class SampleModule(QualityControl):

    def __init__(self, data, attributes):

        QualityControl.__init__(self, data, attributes)

        self.name = "individual"

        self.attributes["modules"][self.name] = {}

        self._set_log()

        stamp("Inititating Sample Module.")

    def _set_log(self):

        self.attributes["modules"][self.name] = {
            "results": {},
            "settings": {},
            "states": {}  # States are other parameters of interest not necessary results or settings.
        }

    def filter_data(self, mind=0.2, recalculate=True):

        """
        Re-write with Pandas
        """

        if mind is None:
            stamp("Returning data without filtering.")
            return self.data, self.attributes

        stamp("Filtering samples with missing data >", mind)
        stamp("Missing data calculated over", len(self.data), "SNPs")

        mind_prop = self._calculate_mind()

        to_remove = mind_prop[mind_prop > mind].index.tolist()

        filtered_data = {}
        for snp, data in self.data.items():
            data["calls"] = [snp_call for i, snp_call in self._iterate_call_indices(data["calls"])
                             if i not in to_remove]
            filtered_data[snp] = data

        attributes = self._adjust_attributes(self.attributes, mind, to_remove)

        percent_removed = format((len(to_remove) / attributes["sample_size"])*100, ".2f")

        stamp("Removed {r} samples out of {t} samples ({p}%)"
              .format(r=len(to_remove), t=attributes["sample_size"], p=percent_removed))

        # Recalculating SNP parameters:

        if recalculate:
            stamp("Recalculating MAF, CALL RATE and HWE for SNPs")
            marker = SNPModule(filtered_data, attributes)
            filtered_data, attributes = marker.get_data(threshold=None)

        return filtered_data, attributes

    @staticmethod
    def _iterate_call_indices(calls):

        for i, snp_call in enumerate(calls):
            yield i, snp_call

    def _calculate_mind(self):

        calls = {snp_id: data["calls"] for snp_id, data in self.data.items()}
        df = pandas.DataFrame(calls)
        mind = (df == self.missing).sum(axis=1)
        mind /= len(self.data)  # Series

        return mind

    def _adjust_attributes(self, attributes, mind, to_remove):

        attributes["modules"][self.name]["results"]["mind"] = {
            "value": mind,
            "removed_samples": len(to_remove)
        }

        attributes["modules"][self.name]["states"]["mind"] = {
            "sample_names_original": attributes["sample_names"],
            "sample_size_original": attributes["sample_size"]
        }

        attributes["sample_names"] = [name for i, name in enumerate(self.sample_names)
                                      if i not in to_remove]

        attributes["sample_size"] -= len(to_remove)

        return attributes

########################################################################################################################


class RedundancyModule(QualityControl):

    def __init__(self, data, attributes, tmp_remove=True):

        QualityControl.__init__(self, data, attributes)

        self.name = "redundancy"

        self.duplicates = {}
        self.retained_duplicates = []
        self.removed_duplicates = []

        self.clusters = {}
        self.retained_sequences = []
        self.removed_sequences = []

        self.identity = 0.95            # Identity used in Cluster Removal

        self.attributes["modules"][self.name] = {}

        self._set_log()

        self.tmp_path = os.path.join(self.out_path, "tmp")
        self.tmp_remove = tmp_remove

    def _set_log(self):

        self.attributes["modules"][self.name] = {
            "results": {"clusters": {
                "before": None,
                "after": None,
                "removed": None
        }, "duplicates": {
                "before": None,
                "after": None,
                "removed": None
        }},
            "settings": {},
            "states": {}  # States are other parameters of interest not necessary results or settings.
        }

    def get_data(self, duplicates=True, clusters=True, redundant=False):

        # Redundant gets retained duplicates and identity clusters
        # else get removed duplicates and clusters

        data = self.data

        self.attributes["modules"][self.name]["settings"] = {
            "duplicates": duplicates,
            "clusters": clusters,
            "redundant": redundant,
            "identity": self.identity
        }

        if redundant:
            duplicate_list = self.retained_duplicates
            cluster_list = self.retained_sequences
        else:
            duplicate_list = self.removed_duplicates
            cluster_list = self.removed_sequences

        # Check if logic is correct:

        if duplicates and clusters:
            filters = [("duplicates", duplicate_list), ("clusters", cluster_list)]
        elif clusters and not duplicates:
            filters = [("clusters", cluster_list)]
        elif duplicates and not clusters:
            filters = [("duplicates", duplicate_list)]
        else:
            filters = list()

        if filters:
            for mode, filter_list in filters:
                before = len(data)
                data = {k: v for (k, v) in data.items() if k not in filter_list}
                after = len(data)

                self._log_filters(mode=mode, before=before, after=after)

                self.messages.get_redundancy_message(mode, before, before-after, after)

        self.attributes["modules"][self.name]["states"] = {
            "duplicates_removed": len(self.removed_duplicates),
            "sequences_removed": len(self.removed_sequences),
            "duplicates_retained": len(self.retained_duplicates),
            "sequences_retained": len(self.retained_sequences),
        }

        return data, self.attributes

    def _log_filters(self, mode, before, after):

        self.attributes["modules"][self.name]["results"][mode] = {
                "before": before,
                "after": after,
                "removed": before - after
        }

    def remove_duplicates(self, selector="maf", target="clone_id", selector_list=None):

        """Search for duplicate clone IDs in SNPs and remove from data. Wrapper. """

        self._find_duplicates(target=target)
        self._select_duplicates(selector=selector, selector_list=selector_list)

    def remove_clusters(self, identity=0.95, target="allele_seq_ref", selector="maf",
                        selector_list=None, cdhit_path=None):

        """ Search for sequence identity clusters and remove them from data. Wrapper. """

        self.identity = identity

        self._find_clusters(target=target, identity=identity, cdhit_path=cdhit_path)
        self._select_clusters(selector=selector, selector_list=selector_list)

    # Private functions for Redundancy Module #

    def _find_clusters(self, target="allele_seq_ref", identity=0.95, word_size=10, description_length=0,
                       cdhit_path=None):

        """
        Clusters the reference allele sequences with CDHIT-EST and parses the clusters for selecting and
        retaining best sequences.

        CD-HIT returns slightly different cluster configurations for each run due to greedy incremental algorithm,
        but little variation observed between runs in the data for P. monodon. Know thyself!

        """

        os.makedirs(self.tmp_path, exist_ok=True)

        fasta_path = self._write_fasta(target)

        cluster_path = self._run_cdhit(fasta_path=fasta_path, identity=identity, word_size=word_size,
                                       description_length=description_length, cdhit_path=cdhit_path)

        self.clusters = self._parse_cdhit(cluster_path)

        if self.tmp_remove:
            shutil.rmtree(self.tmp_path, ignore_errors=True)

    def _select_clusters(self, selector="maf", selector_list=None):

        """ Select best markers from clusters by selector. """

        for cluster, cluster_members in self.clusters.items():
            best_sequence = self._compare_entries(cluster_members, selector=selector, selector_list=selector_list)
            self.retained_sequences.append(best_sequence)
            self.removed_sequences += [member for member in cluster_members if member != best_sequence]

    def _write_fasta(self, target="allele_seq_ref"):

        """ Write fasta file of sequences with SNP IDs for CD-HIT. """

        file_name = os.path.join(self.tmp_path, self.project + "_Seqs")

        seqs = [SeqRecord(Seq(data[target], IUPAC.unambiguous_dna), id=snp_id, name="", description="")
                for snp_id, data in self.data.items()]

        file_name += ".fasta"

        with open(file_name, "w") as fasta_file:
            SeqIO.write(seqs, fasta_file, "fasta")

        return file_name

    def _run_cdhit(self, fasta_path, identity=0.95, word_size=5, description_length=0, cdhit_path=None):

        """ Run CDHIT-EST for sequences, install with sudo apt install cd-hit on Ubuntu """

        self.messages.get_cdhit_message(identity)

        if cdhit_path is None:
            cdhit_path = "cd-hit-est"

        file_name = self.project + "_IdentityClusters_" + str(identity)

        out_file = os.path.join(self.tmp_path, file_name)
        cluster_path = os.path.join(self.tmp_path, file_name + '.clstr')

        with open(os.devnull, "w") as devnull:
            call([cdhit_path, "-i", fasta_path, "-o", out_file, "-c", str(identity), "-n", str(word_size),
                  "-d", str(description_length)], stdout=devnull)

        return cluster_path

    def _parse_cdhit(self, file):

        """
        Parses the CDHIT cluster output and retains only cluster with more than one sequence for selection and
        removal from total SNPs in dictionary: {cluster_id: snp_ids}

        """

        identity_clusters = {}

        with open(file, "r") as clstr_file:
              nrows = len(list(clstr_file))

        with open(file, "r") as clstr_file:
            clstr_reader = csv.reader(clstr_file, delimiter=' ')

            cluster_id = 0
            ids = []

            row_index = 1
            for row in clstr_reader:
                if row[0] == ">Cluster":
                    if len(ids) > 1:
                        identity_clusters[cluster_id] = ids
                    ids = []
                    cluster_id += 1
                    row_index += 1
                else:
                    allele_id = self._find_between(row[1], ">", "...")
                    ids.append(allele_id)
                    # EOF
                    if row_index == nrows:
                        if len(ids) > 1:
                            identity_clusters[cluster_id] = ids
                    row_index += 1

        return identity_clusters

    @staticmethod
    def _find_between(s, first, last):

        """Finds the substring between two strings."""

        try:
            start = s.index(first) + len(first)
            end = s.index(last, start)
            return s[start:end]
        except ValueError:
            return ""

    def _find_duplicates(self, target="clone_id"):

        """ Count duplicates in target category (usually clone ID) """

        clone_counts = {}

        for k, v in self.data.items():

            clone_id = v[target]
            if clone_id not in clone_counts.keys():
                clone_counts[clone_id] = {"count": 1, "allele_ids": [k]}
            else:
                clone_counts[clone_id]["count"] += 1
                clone_counts[clone_id]["allele_ids"].append(k)

        self.duplicates = {k: v for (k, v) in clone_counts.items() if v["count"] > 1}

    def _select_duplicates(self, selector="maf", selector_list=None):

        """ Select best SNP from duplicate clusters. """

        for clone, clone_data in self.duplicates.items():
            best_clone = self._compare_entries(clone_data["allele_ids"], selector=selector, selector_list=selector_list)
            self.retained_duplicates.append(best_clone)
            self.removed_duplicates += [marker for marker in clone_data["allele_ids"] if marker != best_clone]

    def _compare_entries(self, ids, selector="maf", selector_list=None):

        """
        Gets data from dictionary for each duplicate SNP according to 'selector'
        and returns the allele identification of the best entry.

        Selector list currently sorts descending, that is all selector values must be ranked highest value ("best") -
        this is the case for MAF, Call Rate, Rep, Read Counts ...

        Later rank the data by QC Score.

        """

        if selector_list is None:
            entries_stats = [[i, self.data[i][selector]] for i in ids]
            entries_ranked = sorted(entries_stats, key=operator.itemgetter(1), reverse=True)
        else:
            entries_stats = [[i] + [self.data[i][selector] for selector in selector_list] for i in ids]
            entries_ranked = sorted(entries_stats, key=operator.itemgetter(*[i for i in range(1, len(selector_list)+1)]),
                                    reverse=True)

        return entries_ranked[0][0]

########################################################################################################################


class CombinationModule:

    """ Combinator module for MarkerModule. Assess and visualize combinations of parameters for QC. """

    def __init__(self, marker_module):

        if not marker_module.filters:
            ValueError("Data must have been assessed and filtered with MarkerModule.")

        self.filters = marker_module.filters
        self.data = marker_module.data

    def get_matrix(self, parameter_one, parameter_two, values_one, values_two):

        """"
        Pairwise analysis of two values given two parameters and value vectors, outputs data for visualization of
        matrix in Excel and R.

        """

        result_matrix = [[parameter_one + "/" + parameter_two] + values_one]
        r_matrix = []

        try:
            for value_y in values_two:
                result_row = [value_y]
                filtered_y = self.filters[parameter_two][value_y]
                for value_x in values_one:
                    filtered_x = self.filters[parameter_one][value_x]
                    filter_combined = set(filtered_x + filtered_y)
                    r_matrix += [[str(value_y), str(value_x), len(self.data) - len(filter_combined), 100, 100]]
                    result_row.append(len(self.data) - len(filter_combined))        # Number of retained SNPs
                result_matrix.append(result_row)

            return result_matrix, r_matrix

        except KeyError:
            "Can't find parameter or specified value in marker filters, please use MarkerModule"


class SNPModule(QualityControl):

    """ Analysis module for markers, calculate parameters and filter SNPs. """

    def __init__(self, data, attributes):

        QualityControl.__init__(self, data, attributes)

        self.name = "snp"

        self.filters = {}                   # {"maf" : {0.5 : [ "SNP1" , "SNP2" ...] ...} ...}

        self._calculate_parameters()

        self.attributes["modules"][self.name] = {}

        self._set_log()

    def _set_log(self):

        self.attributes["modules"][self.name] = {
            "results": {},
            "settings": {},
            "states": {}  # States are other parameters of interest not necessary results or settings.
        }

    def _log_filters(self, parameter, value, before, after):

        self.attributes["modules"][self.name]["results"][parameter] = {
                "value": value,
                "before": before,
                "after": after,
                "removed": before - after
        }

    def filter_data(self, thresholds, parameter="maf", comparison="<="):

        """
        Filter data by a list of thresholds and parameter / comparsion.
        Filtered marker IDs are stored in filter attribute and can be returned with .get_data().
        """

        if comparison not in ["<=", ">=", "=="]:
            raise ValueError("Comparison must be one of: <=, >=, ==")

        for threshold in thresholds:
            if threshold is not None:
                if comparison == "<=":
                    filtered = [k for (k, v) in self.data.items() if v[parameter] <= threshold]
                elif comparison == ">=":
                    filtered = [k for (k, v) in self.data.items() if v[parameter] >= threshold]
                else:
                    filtered = [k for (k, v) in self.data.items() if v[parameter] == threshold]

                try:
                    self.filters[parameter][threshold] = filtered
                except KeyError:
                    self.filters[parameter] = {threshold: filtered}

    def get_data(self, threshold=None, parameter="maf", multiple=None):

        """
        Multiple is a dictionary of parameter keys and parameter values (e.g. {"maf": 0.5, "hwe": 0.0001}) returning
        data filtered by multiple parameters.

        """

        if multiple is not None:

            self.attributes["modules"][self.name]["settings"] = {
                "parameters": multiple  # Dictionary
            }

            data = self.data
            for parameter, threshold in multiple:
                before = len(data)

                if threshold is not None:
                    data = {k: v for (k, v) in data.items() if k not in self.filters[parameter][threshold]}

                after = len(data)

                self._log_filters(parameter=parameter, before=before, after=after, value=threshold)

                self.messages.get_filter_message(parameter, threshold, before, before-after, after)

            return data, self.attributes

        if threshold is None:
            return self.data, self.attributes
        else:

            data = {k: v for (k, v) in self.data.items() if k not in self.filters[parameter][threshold]}

            self.attributes["modules"][self.name]["settings"] = {
                "parameter": parameter,
                "value": threshold
            }

            self._log_filters(parameter=parameter, before=len(self.data), after=len(data), value=threshold)

            self.messages.get_filter_message(parameter, threshold, len(self.data), len(self.data)-len(data), len(data))

            return data, self.attributes

    # Private functions for parameter calculations

    def _calculate_maf(self, snp, genotypes):

        """
        Calculates minor allele frequency for a single SNP. Pass a one-row format list (zip on double-row)
        of allele calls. Returns the minimum allele frequency for processing.
        """

        if self.sample_size != len(genotypes):
            print("Warning: Number of samples does not correspond number of allele calls.")

        adjusted_samples = self.sample_size - genotypes.count(self.missing)

        het_count = genotypes.count(self.heterozygous)

        try:
            freq_allele_one = (genotypes.count(self.homozygous_major) + (het_count/2)) / adjusted_samples
            freq_allele_two = (genotypes.count(self.homozygous_minor) + (het_count/2)) / adjusted_samples
        except ZeroDivisionError:
            #print("Detected complete missing data in SNP:", snp)
            return 0

        return min(freq_allele_one, freq_allele_two)

    def _calculate_hwe(self, snp, genotypes):

        """
        Calculates p-value for HWE using ChiSquare statistic: remove missing, get observed counts, get observed
        frequencies, get expected counts, calculate test values using (O-E)**2 / E and return ChiSquare probability
        with 1 degree of Freedom (bi-allelic SNP).

        """

        adjusted_samples = self.sample_size - genotypes.count(self.missing)
        hetero_obs, major_obs, minor_obs = self._get_observed(genotypes)

        try:
            p = (major_obs + (hetero_obs/2)) / adjusted_samples
            q = (minor_obs + (hetero_obs/2)) / adjusted_samples
        except ZeroDivisionError:
            #print("Detected complete missing data in SNP:", snp)
            return 0

        if (p + q) != 1:
            ValueError("Sum of observed allele frequencies (p + q) does not equal one.")

        hetero_exp, major_exp, minor_exp = self._get_expected(p, q, adjusted_samples)

        try:
            hetero_test = ((hetero_obs-hetero_exp)**2)/hetero_exp
            major_test = ((major_obs-major_exp)**2)/major_exp
            minor_test = ((minor_obs-minor_exp)**2)/minor_exp
        except ZeroDivisionError:
            return 0

        return stats.chisqprob(sum([hetero_test, major_test, minor_test]), 1)

    def _get_observed(self, genotypes):

        """" Get observed counts in the genotype for each SNP """

        return genotypes.count(self.heterozygous), genotypes.count(self.homozygous_major),\
               genotypes.count(self.homozygous_minor)

    @staticmethod
    def _get_expected(p, q, adjusted_samples):

        """ Get expected counts under HWE """

        return adjusted_samples*(2*p*q), adjusted_samples*(p**2), adjusted_samples*(q**2)

    def _calculate_call(self, genotypes):

        """ Calculates call rate across samples for a single SNP. """

        return 1 - (genotypes.count(self.missing)/self.sample_size)

    def _calculate_parameters(self):

        """
        Calculate minor allele frequency, call rate and chi-square probability for HWE for each SNP and decorate
        the data dictionary for statistics across SNPs.
        """

        for snp, data in self.data.items():

            self.data[snp]["maf"] = self._calculate_maf(snp, data["calls"])
            self.data[snp]["call_rate"] = self._calculate_call(data["calls"])
            self.data[snp]["hwe"] = self._calculate_hwe(snp, data["calls"])

########################################################################################################################



from distutils.core import setup


def readme():
    with open('README.md', 'r') as file:
        file.read()

setup(name='dartqc',
      version='0.1.4',
      description='Quality control for SNP data from Diversity Array Technologies (DArT)',
      url='http://github.com/esteinig/dartqc',
      download_url='http://github.com/esteinig/dartqc/archive/0.1.4.tar.gz',
      author='Eike J. Steinig',
      author_email='eikejoachim.steinig@my.jcu.edu.au',
      license='MIT',
      packages=["dartqc"],
      scripts=['bin/dartqc'],
      package_dir={"dartqc": "dartqc"},
      package_data={"dartqc": ["env/dartqc.yaml"]})



import os
import json
import pandas
import operator

from dartqc.DartUtils import stamp


class DartPreparator:

    """
    Class for guessing specifications of input file and converting to input for DartReader

    limit: number of beginning rows to read that likely contain the meta data, usually no more than 30

    """

    def __init__(self, file_path, output_name="dartqc", output_path=os.getcwd(), excel_sheet=""):

        self.config = {
            "clone_column": [
                "CloneID"
            ],
            "allele_column": [
                "AlleleID"
            ],
            "sequence_column": [
                "AlleleSequence"
            ],
            "replication_column": [
                "RepAvg"
            ]
        }

        self.file_path = file_path
        self.output_name = output_name
        self.output_path = output_path

        self.excel_sheet = excel_sheet

        self.limit = 30

        self.top = None

        self.data_row = 0
        self.sample_row = 0
        self.data_column = 0

        self.header = None
        self.scheme = dict()

        if self.excel_sheet:
            self._convert_excel()

        self._read_csv()
        self._get_row_indices()
        self._get_column_indices()
        self._reindex()
        self._write_scheme()

    def _convert_excel(self):

        stamp("Converting from Excel")
        stamp("File is", self.file_path)
        stamp("Sheet is", self.excel_sheet)
        data_xls = pandas.read_excel(self.file_path, self.excel_sheet, index_col=None)
        name, ext = os.path.splitext(os.path.basename(self.file_path))
        outfile = os.path.join(self.output_path, name + ".csv")

        stamp("Writing to file", outfile)
        data_xls.to_csv(outfile, encoding='utf-8', index=False)

        self.file_path = outfile

    def _read_csv(self):

        stamp("Loading file", self.file_path)

        self.top = pandas.read_csv(self.file_path, header=None, nrows=30)

    def _get_row_indices(self):

        """

        This function gets the row indices for samples and data, as well as the header, assuming:

            - header row begins after rows starting with "*"
            - data row starts after header row
            - samples are specified in the header row (above calls or raw counts)

        """

        stamp("Guessing data configuration:")

        for i, row in self.top.iterrows():
            if row[0] != "*":
                self.header = row
                self.sample_row = i
                self.data_row = i+1
                self.scheme["sample_row"] = self.sample_row
                self.scheme["data_row"] = self.data_row
                break

    def _get_column_indices(self):

        """

        Get column indices for key words defined in the config file (excel_scheme.json), also get
        column indices for start of data assuming:

            - data starts in first column after placeholder ("*")
            - guessed from the row above the header (self.sample_row-1)

        """

        i = 0
        for col in self.top:
            data = self.top[col]
            if data[self.sample_row-1] != "*":
                self.scheme["data_column"] = i
                break
            i += 1

        for column, aliases in self.config.items():
            column_indices = [pandas.Index(self.header).get_loc(header) for header in aliases]

            if len(column_indices) < 1:
                raise ValueError("Could not find column header", column, "in dataset header. "
                                 "Please provide a different alias in dartQC/schemes/excel_scheme.json.")
            elif len(column_indices) >= 2:
                raise ValueError("Found more than one column header out of:", aliases, "make sure column "
                                 "headers are not repeated in data. You can change aliases that this command is "
                                 "looking for in dartQC/schemes/excel_scheme.json")
            else:
                self.scheme[column] = column_indices[0]

    def _reindex(self):

        """ Reindex values for non-pythonic input to DartReader (better for users) """

        self.scheme = {k: int(v+1) for k, v in self.scheme.items()}

        for k, v in sorted(self.scheme.items(), key=operator.itemgetter(1)):
            stamp(k, "=", v)

        stamp("Please check these values in your data to ensure correct input for DartQC.")

    def _write_scheme(self):

        if self.output_name is None:
            name, ext = os.path.splitext(os.path.basename(self.file_path))
            file_name = name + "_scheme.json"
        else:
            file_name = self.output_name + "_scheme.json"

        out_file = os.path.join(self.output_path, file_name)

        stamp("Writing scheme to:", out_file)

        with open(out_file, "w") as outfile:
            json.dump(self.scheme, outfile, indent=4)




import os
import csv
import json


class DartReader:

    """Class for reading raw calls."""

    def __init__(self):

        self.project = "Monodon"
        self.verbose = True

        self.out_path = ""

        # Parsing raw data

        self.raw_file = ''              # File name with raw data
        self.data = {}                  # Holds initial unfiltered data
        self.header = []                # Holds the lines before the actual header for statistics and data

        self.sample_names = []
        self.sample_size = 0
        self.snp_number = 0

        self.format = "double"

        # Row numbers (non-pythonic) in Excel Spreadsheet

        self._data_row = 7              # Start of Sequences / Data
        self._sample_row = 5            # Sample Identification
        self._pop_row = 0               # Sample Populations

        # Column numbers (non-pythonic) in Excel Spreadsheet

        self._id = 1
        self._clone = 2
        self._seq = 3
        self._snp = 4
        self._snp_position = 5
        self._call_rate_dart = 6
        self._one_ratio_ref = 7
        self._one_ratio_snp = 8
        self._freq_homozygous_ref = 9
        self._freq_homozygous_snp = 10
        self._freq_heterozygous = 11
        self._pic_ref = 12
        self._pic_snp = 13
        self._pic_average = 14
        self._read_count_ref = 15
        self._read_count_snp = 16
        self._rep_average = 17
        self._call = 18
        self._sample_column = 18

        # Meta Data by Individuals

        self.meta = {}

        self._id_meta = 1
        self._pop_meta = 2

        self.split_clone = False
        self.clone_split = "|"

        # Encoding Scheme

        self.homozygous_major = ("1", "0")
        self.homozygous_minor = ("0", "1")
        self.heterozygous = ("1", "1")
        self.missing = ("-", "-")

        self._dart_qc_encoding = {self.missing: "-", self.heterozygous: "0", self.homozygous_minor: "1",
                                  self.homozygous_major: "2"}

    def set_options(self, project="DartQC", verbose=True, out_path=os.getcwd(), homozygous_major=("1", "0"),
                    homozygous_minor=("0", "1"), heterozygous=("1", "1"), missing=("-", "-"), pop_row=0, sample_row=5,
                    data_start_row=7, id_col=1, clone_col=2, seq_col=3, snp_col=4, snp_position_col=5, call_rate_col=6,
                    one_ratio_ref_col=7, one_ratio_snp_col=8, freq_homozygous_ref_col=9, freq_homozygous_snp_col=10,
                    freq_heterozygous_col=11, pic_ref_col=12, pic_snp_col=13, pic_average=14, read_count_ref_col=15,
                    read_count_snp_col=16, rep_average_col=17, call_start_col=18, sample_start_col=18, clone_split="|",
                    split_clone=False, scheme=None):

        self.project = project
        self.verbose = verbose

        self.out_path = out_path

        self._data_row = data_start_row
        self._sample_row = sample_row
        self._pop_row = pop_row

        self.homozygous_major = homozygous_major
        self.homozygous_minor = homozygous_minor
        self.heterozygous = heterozygous
        self.missing = missing

        self._id = id_col
        self._clone = clone_col
        self._seq = seq_col
        self._snp = snp_col
        self._snp_position = snp_position_col
        self._call_rate_dart = call_rate_col
        self._one_ratio_ref = one_ratio_ref_col
        self._one_ratio_snp = one_ratio_snp_col
        self._freq_homozygous_ref = freq_homozygous_ref_col
        self._freq_homozygous_snp = freq_homozygous_snp_col
        self._freq_heterozygous = freq_heterozygous_col
        self._pic_ref = pic_ref_col
        self._pic_snp = pic_snp_col
        self._pic_average = pic_average
        self._read_count_ref = read_count_ref_col
        self._read_count_snp = read_count_snp_col
        self._rep_average = rep_average_col
        self._call = call_start_col
        self._sample_column = sample_start_col

        self.split_clone = split_clone
        self.clone_split = clone_split

        self._dart_qc_encoding = {self.missing: "-", self.heterozygous: "0", self.homozygous_minor: "1",
                                  self.homozygous_major: "2"}

        if scheme is not None:
            # Only basic supported for now.
            with open(scheme, "r") as infile:
                config = json.load(infile)

                self._id = config["allele_column"]
                self._clone = config["clone_column"]
                self._seq = config["sequence_column"]
                self._rep_average = config["replication_column"]
                self._call = config["data_column"]
                self._sample_column = config["data_column"]
                self._data_row = config["data_row"]
                self._sample_row = config["sample_row"]

    def read_pops(self, file, sep=','):

        """ Read file with header and two columns: 1 - ID, 2 - Population. ID must be the same as in Data. """

        meta_head = []

        with open(file, 'r') as infile:
            reader = csv.reader(infile, delimiter=sep)
            for row in reader:
                if meta_head:
                    self.meta[row[self._id_meta-1]] = row[self._pop_meta-1]
                else:
                    meta_head = row

    def read_double_row(self, file, encode=True, split_char="|", numeric=False, basic=True):

        """"Read data in double row format"""

        self.raw_file = file
        self.format = "double"
        self.clone_split = split_char

        with open(file, 'r') as data_file:
            reader = csv.reader(data_file)

            row_index = 1  # Non-pythonic numbering for Excel Users
            snp_count = 0

            allele_id = None
            allele_index = 1

            pops = []

            for row in reader:

                if row_index <= self._data_row-2:  # Don't include description header
                    self.header.append(row)

                if row_index == self._sample_row:
                    self.sample_names = row[self._sample_column-1:]
                    self.sample_size = len(self.sample_names)

                if row_index == self._pop_row:
                    pops = row[self._sample_column-1:]

                # Data Rows, read from specified row and only if it contains data in at least one field (remove empties)
                if row_index >= self._data_row and any(row):

                    # Get reduced data by unique allele ID in double Rows (K: Allele ID, V: Data)
                    # Implement Error checks for conformity between both alleles: SNP Position, Number of Individuals
                    if allele_index == 1:

                        allele_id = row[self._id-1]
                        clone_id = row[self._clone-1]

                        call_1 = row[self._call-1:]

                        if numeric:
                            call_1 = [int(call) for call in call_1]

                        if self.split_clone:
                            clone_id = clone_id.split(self.clone_split)[0]

                        if basic:
                            entry = {"allele_id": allele_id,
                                     "clone_id": clone_id,
                                     "allele_seq_ref": row[self._seq-1],
                                     "rep_average": float(row[self._rep_average-1]),
                                     "calls": [call_1]}  # Add allele calls 1
                        else:

                            entry = {"allele_id": allele_id,
                                     "clone_id": clone_id,
                                     "allele_seq_ref": row[self._seq-1],
                                     "snp_position": row[self._snp_position-1],
                                     "call_rate_dart": row[self._call_rate_dart-1],
                                     "one_ratio_ref": row[self._one_ratio_ref-1],
                                     "one_ratio_snp": row[self._one_ratio_snp-1],
                                     "freq_homozygous_ref": row[self._freq_homozygous_ref-1],
                                     "freq_homozygous_snp": row[self._freq_homozygous_snp-1],
                                     "freq_heterozygous": row[self._freq_heterozygous-1],
                                     "pic_ref": row[self._pic_ref-1],
                                     "pic_snp": row[self._pic_snp-1],
                                     "pic_average": row[self._pic_average-1],
                                     "read_count_ref": float(row[self._read_count_ref-1]),
                                     "read_count_snp": float(row[self._read_count_snp-1]),
                                     "rep_average": float(row[self._rep_average-1]),
                                     "calls": [call_1]}  # Add allele calls 1

                        self.data[allele_id] = entry

                        snp_count += 1
                        allele_index = 2
                    else:
                        # Add sequence and allele calls 2:

                        call_2 = row[self._call-1:]

                        if numeric:
                            call_2 = [int(call) for call in call_2]

                        self.data[allele_id]["calls"].append(call_2)
                        self.data[allele_id]["allele_seq_snp"] = row[self._seq-1]
                        self.data[allele_id]["snp"] = row[self._snp-1]

                        if len(self.data[allele_id]["calls"]) != 2:
                            raise(ValueError("Error. Genotype of", allele_id, "does not contain two alleles.",
                                  "Check if starting row for data is correctly specified."))
                        if encode:
                            self.data[allele_id]["calls"] = self._encode_dart(self.data[allele_id]["calls"])
                        else:
                            self.data[allele_id]["calls"] = list(zip(self.data[allele_id]["calls"][0],
                                                                     self.data[allele_id]["calls"][1]))

                        self.snp_number += 1
                        allele_index = 1

                row_index += 1

        # Check if reader picked up populations, if not generate generic names:
        if not pops:
            pops = ["Pop" for i in range(self.sample_size)]

        self.meta = dict(zip(self.sample_names, pops))

    def read_single_row(self, file):

        self.raw_file = file
        self.format = "single"

        pass

    def read_json(self, data_file, attribute_file):

        with open(data_file) as data_in:
            data = json.load(data_in)

        with open(attribute_file) as attr_in:
            attributes = json.load(attr_in)

        return data, attributes

    def get_data(self):

        attributes = {

            "project": self.project,
            "sample_size": self.sample_size,
            "sample_names": self.sample_names,
            "pops": self.meta,
            "missing": self._dart_qc_encoding[self.missing],
            "heterozygous": self._dart_qc_encoding[self.heterozygous],
            "homozygous_minor": self._dart_qc_encoding[self.homozygous_minor],
            "homozygous_major": self._dart_qc_encoding[self.homozygous_major],
            "out_path": self.out_path,
            "file": self.raw_file,
            "snps": len(self.data),
            "modules": {}

        }

        return self.data, attributes

    def _encode_dart(self, calls):

        if self.format == "double":
            calls = zip(calls[0], calls[1])

        return [self._dart_qc_encoding[snp_call] for snp_call in calls]